(** 

This file describes the syntax and semantics of a
lambda calculus with references. 

Author: Arthur Charguéraud.
License: MIT.

*)

Set Implicit Arguments.
Require Export LibCore ModelState.


(************************************************************)
(* * Source language syntax *)

(** Representation of variables and locations *)

Definition var := nat.
Definition loc := nat.
Definition null : loc := 0%nat.

Global Opaque loc.


(** Syntax of the source language *)

Inductive prim : Type :=
  | prim_ref : prim
  | prim_get : prim
  | prim_set : prim.

Inductive val : Type := 
  | val_unit : val
  | val_int : int -> val
  | val_pair : val -> val -> val
  | val_var : var -> val
  | val_loc : loc -> val
  | val_prim : prim -> val
  | val_fix : var -> var -> trm -> val

with trm : Type :=
  | trm_val : val -> trm
  | trm_if : val -> trm -> trm -> trm
  | trm_let : var -> trm -> trm -> trm
  | trm_app : val -> val -> trm.


(** Definition of capture-avoiding substitution *)

Fixpoint subst_val (y : var) (w : val) (v : val) : val :=
  match v with
  | val_var x => If x = y then w else v
  | val_fix f x t => If (x = y \/ f = y) then v else val_fix f x (subst_trm y w t) 
  | _ => v 
  end

with subst_trm (y : var) (w : val) (t : trm) : trm :=
  match t with
  | trm_val v => trm_val (subst_val y w v)
  | trm_if v t1 t2 => trm_if (subst_val y w v) (subst_trm y w t1) (subst_trm y w t2)
  | trm_let x t1 t2 => trm_let x (subst_trm y w t1) (If x = y then t2 else subst_trm y w t2)
  | trm_app v1 v2 => trm_app (subst_val y w v1) (subst_val y w v2)
  end.



(************************************************************)
(* * Source language semantics *)

(*------------------------------------------------------------------*)
(** Big-step evaluation *)

Section Red.

Definition state := state loc val.

Local Open Scope state_scope.

Coercion val_prim : prim >-> val.
Coercion trm_val : val >-> trm.
Coercion trm_app : val >-> Funclass.

Implicit Types t : trm.
Implicit Types v : val.

Inductive red : state -> trm -> state -> val -> Prop :=
  | red_val : forall m v,
      red m v m v
  | red_if : forall m1 m2 v r t1 t2,
      red m1 (If v = val_int 0 then t2 else t1) m2 r ->
      red m1 (trm_if v t1 t2) m2 r
  | red_let : forall m1 m2 m3 x t1 t2 v1 r,
      red m1 t1 m2 v1 ->
      red m2 (subst_trm x v1 t2) m3 r ->
      red m1 (trm_let x t1 t2) m3 r
  | red_app : forall m1 m2 v1 v2 f x t r,
      v1 = val_fix f x t ->
      red m1 (subst_trm f v1 (subst_trm x v2 t)) m2 r ->
      red m1 (trm_app v1 v2) m2 r
  | red_ref : forall ma mb v l,  
      l <> null ->
      mb = (state_single l v) ->
      \# ma mb ->
      red ma (prim_ref v) (mb \+ ma) (val_loc l)
  | red_get : forall m l v,
      state_data m l = Some v ->
      red m (prim_get (val_loc l)) m v
  | red_set : forall m m' l v,
      m' = state_update m l v ->
      red m (prim_set (val_pair (val_loc l) v)) m' val_unit.

End Red.


(*------------------------------------------------------------------*)
(* ** Tactic [state_red] for proving [red] goals modulo 
      equalities between states *)

Ltac state_red_base tt ::=
  match goal with H: red _ ?t _ _ |- red _ ?t _ _ =>
    applys_eq H 2 4; try state_eq end.

