(** 

This file formalizes "Separation Logic with Time Credits",
following the presentation by Arthur Charguéraud and 
François Pottier described in the following papers:

- Machine-Checked Verification of the Correctness and Amortized 
  Complexity of an Efficient Union-Find Implementation (ITP'15)

- Verifying the Correctness and Amortized Complexity of a Union-Find 
  Implementation in Separation Logic with Time Credits
  (Draft journal paper)

This file contains:
- a definition of heaps as finite maps from location to values
  paired with a natural number,
- an instantiation of the functor from the file ModelSepFunctor.v,
- a definition of triples,
- statement and proofs of SL reasoning rules.

Author: Arthur Charguéraud.
License: MIT.

*)

Set Implicit Arguments.
Require Export ModelLambda ModelSepFunctor.
Open Scope state_scope.



(********************************************************************)
(* * Semantics *)

(*------------------------------------------------------------------*)
(* ** Big-step evaluation with counting of the number of beta reductions
    (used by the formalization of Separation Logic with time credits) *)

Section Redn.
Local Open Scope nat_scope.
Local Open Scope state_scope.

Inductive red : nat -> state -> trm -> state -> val -> Prop :=
  | red_val : forall m v,
      red 0 m (trm_val v) m v
  | red_if : forall n m1 m2 v r t1 t2,
      red n m1 (If v = val_int 0 then t2 else t1) m2 r ->
      red n m1 (trm_if v t1 t2) m2 r
  | red_let : forall n1 n2 m1 m2 m3 x t1 t2 v1 r,
      red n1 m1 t1 m2 v1 ->
      red n2 m2 (subst_trm x v1 t2) m3 r ->
      red (n1+n2) m1 (trm_let x t1 t2) m3 r
  | red_app : forall n m1 m2 v1 v2 f x t r,
      v1 = val_fix f x t ->
      red n m1 (subst_trm f v1 (subst_trm x v2 t)) m2 r ->
      red (S n) m1 (trm_app v1 v2) m2 r
  | red_ref : forall ma mb v l,
      mb = (state_single l v) ->
      \# ma mb ->
      red 0 ma (prim_ref v) (mb \+ ma) (val_loc l)
  | red_get : forall m l v,
      state_data m l = Some v ->
      red 0 m (prim_get (val_loc l)) m v
  | red_set : forall m m' l v,
      m' = state_update m l v ->
      red 0 m (prim_set (val_pair (val_loc l) v)) m' val_unit.

End Redn.


(********************************************************************)
(* * Construction of core of the logic *)

Module SepCreditsCore.

(*------------------------------------------------------------------*)
(* ** Representation of credits *)

(** Representation of credits *)

Definition credits : Type := nat.

(** Zero and one credits *)

Definition credits_zero : credits := 0%nat.
Definition credits_one : credits := 1%nat.


(*------------------------------------------------------------------*)
(* ** Types *)

Definition heap : Type := (state * credits)%type.

Definition hprop := heap -> Prop.


(*------------------------------------------------------------------*)
(* ** Operations on heaps *)

(** Empty heap *)

Definition heap_empty : heap :=
  (state_empty, 0%nat). 

(** Projections *)

Coercion heap_state (h:heap) : state :=
  match h with (m,c) => m end.

Definition heap_credits (h:heap) : credits :=
  match h with (m,c) => c end.

Notation "h '^s'" := (heap_state h)
   (at level 9, format "h '^s'") : heap_scope.

Notation "h '^c'" := (heap_credits h)
   (at level 9, format "h '^c'") : heap_scope.

Open Scope heap_scope.

(** Disjoint heaps *)

Definition heap_disjoint (h1 h2 : heap) : Prop :=
  \# (h1^s) (h2^s).

Notation "\# h1 h2" := (heap_disjoint h1 h2)
  (at level 40, h1 at level 0, h2 at level 0, no associativity) : heap_scope.

(** Union of heaps *)

Definition heap_union (h1 h2 : heap) : heap :=
   (h1^s \+ h2^s, (h1^c + h2^c)%nat).

Notation "h1 \u h2" := (heap_union h1 h2)
   (at level 51, right associativity) : heap_scope.


(*------------------------------------------------------------------*)
(* ** Operators *)

Definition hprop_empty : hprop :=
  fun h => h = heap_empty.

Definition hprop_star (H1 H2 : hprop) : hprop := 
  fun h => exists h1 h2, 
              H1 h1 
           /\ H2 h2 
           /\ heap_disjoint h1 h2 
           /\ h = h1 \u h2.

Definition hprop_exists A (J : A -> hprop) : hprop := 
  fun h => exists x, J x h.

Definition hprop_pure (P:Prop) : hprop := 
  hprop_exists (fun (p:P) => hprop_empty).

Definition hprop_gc := 
  hprop_exists (fun (H:hprop) => H).


(*------------------------------------------------------------------*)
(* ** Notation *)

Notation "\[]" := (hprop_empty) 
  (at level 0) : heap_scope.
Notation "\[ P ]" := (hprop_pure P) 
  (at level 0, P at level 99) : heap_scope.
Notation "H1 '\*' H2" := (hprop_star H1 H2)
  (at level 41, right associativity) : heap_scope.
Notation "\GC" := (hprop_gc) : heap_scope. 
Open Scope heap_scope.


(*------------------------------------------------------------------*)
(* ** Tactic for automation *)

Hint Extern 1 (_ = _ :> heap) => state_eq.

Lemma heap_disjoint_def : forall h1 h2,
  heap_disjoint h1 h2 = state_disjoint (h1^s) (h2^s).
Proof using. auto. Qed.

Hint Rewrite heap_disjoint_def : rew_disjoint.

Tactic Notation "state_disjoint_pre" :=
  subst; rew_disjoint; jauto_set.

Hint Extern 1 (\# _ _) => state_disjoint_pre.

Ltac math_0 ::= unfolds credits.


(*------------------------------------------------------------------*)
(* ** Equalities on [heap] *)

Lemma heap_eq : forall h1 h2,
  (h1^s = h2^s /\ h1^c = h2^c) -> h1 = h2.
Proof using.
  intros (s1,n1) (s2,n2) (M1&M2). simpls. subst. fequals.
Qed.

Lemma heap_eq_forward : forall h1 h2,
  h1 = h2 ->
  h1^s = h2^s /\ h1^c = h2^c.
Proof using. intros (s1,n1) (s2,n2) M. inverts~ M. Qed.


(*------------------------------------------------------------------*)
(* ** Properties of [heap_disjoint] *)

Lemma heap_disjoint_sym : forall h1 h2,
  \# h1 h2 -> \# h2 h1.
Proof using.
  intros [m1 n1] [m2 n2] H. simpls.
  hint state_disjoint_sym. autos*.
Qed.

Lemma heap_disjoint_comm : forall h1 h2,
  \# h1 h2 = \# h2 h1.
Proof using. 
  intros [m1 n1] [m2 n2]. simpls.
  hint state_disjoint_sym. extens*.
Qed. 

Lemma heap_disjoint_empty_l : forall h,
  \# heap_empty h.
Proof using. intros [m n]. hint state_disjoint_empty_l. hnfs*. Qed.

Lemma heap_disjoint_empty_r : forall h,
  \# h heap_empty.
Proof using. intros [m n]. hint state_disjoint_empty_r. hnfs*. Qed.

Lemma heap_disjoint_union_eq_r : forall h1 h2 h3,
  \# h1 (h2 \u h3) = (\# h1 h2 /\ \# h1 h3).
Proof using.
  intros [m1 n1] [m2 n2] [m3 n3].
  unfolds heap_disjoint, heap_union. simpls.
  rewrite state_disjoint_union_eq_r. extens*.
Qed.

Lemma heap_disjoint_union_eq_l : forall h1 h2 h3,
  \# (h2 \u h3) h1 = (\# h1 h2 /\ \# h1 h3).
Proof using.
  intros. rewrite heap_disjoint_comm. 
  apply heap_disjoint_union_eq_r.
Qed.

Hint Resolve 
   heap_disjoint_sym 
   heap_disjoint_empty_l heap_disjoint_empty_r
   heap_disjoint_union_eq_l heap_disjoint_union_eq_r.

Tactic Notation "rew_disjoint" :=
  autorewrite with rew_disjoint in *.
Tactic Notation "rew_disjoint" "*" :=
  rew_disjoint; auto_star.


(*------------------------------------------------------------------*)
(* ** Properties of [heap_union] *)

Lemma heap_union_comm : forall h1 h2,
  \# h1 h2 ->
  h1 \u h2 = h2 \u h1.
Proof using.
  intros [m1 n1] [m2 n2] H. unfold heap_union.
  simpl. fequals. state_eq. math.
Qed.

Lemma heap_union_assoc : forall h1 h2 h3,
  (h1 \u h2) \u h3 = h1 \u (h2 \u h3).
Proof using.
  intros [m1 n1] [m2 n2] [m3 n3]. unfolds heap_union.
  simpl. fequals. state_eq. math.
Qed.

Lemma heap_union_empty_l : forall h,
  heap_empty \u h = h.
Proof using. 
  intros [m n]. unfold heap_union, heap_empty. simpl.
  fequals. apply~ state_union_empty_l.
Qed.

Lemma heap_union_empty_r : forall h,
  h \u heap_empty = h.
Proof using. 
  intros. rewrite~ heap_union_comm. apply~ heap_union_empty_l.
Qed.

Lemma heap_union_state : forall h1 h2,
  heap_state (h1 \u h2) = (heap_state h1) \+ (heap_state h2).
Proof using. intros (m1&n1) (m2&n2). auto. Qed.

Lemma heap_union_credits : forall h1 h2,
  heap_credits (h1 \u h2) = (heap_credits h1 + heap_credits h2)%nat.
Proof using. intros (m1&n1) (m2&n2). auto. Qed.

Hint Resolve heap_union_comm
   heap_union_empty_l heap_union_empty_r.

Hint Rewrite heap_union_state : rew_disjoint.

Hint Rewrite heap_union_state : rew_state.

Hint Rewrite 
  heap_union_empty_l heap_union_empty_r
  heap_union_state heap_union_credits : rew_heap.

Tactic Notation "rew_heap" :=
  autorewrite with rew_heap.
Tactic Notation "rew_heap" "~" :=
  rew_heap; auto_tilde.
Tactic Notation "rew_heap" "in" hyp(H) :=
  autorewrite with rew_heap in H.
Tactic Notation "rew_heap" "~" "in" hyp(H) :=
  rew_heap in H; auto_tilde.
Tactic Notation "rew_heap" "in" "*" :=
  autorewrite with rew_heap in *.
Tactic Notation "rew_heap" "~" "in" "*" :=
  rew_heap in *; auto_tilde.

Ltac heap_eq :=
  solve [ rew_heap; subst; auto ].


(*------------------------------------------------------------------*)
(* ** Properties of empty *)

Lemma hprop_empty_prove : 
  \[] heap_empty.
Proof using. hnfs~. Qed.

Lemma hprop_empty_inv : forall h,
  \[] h ->
  h = heap_empty.
Proof using. auto. Qed.


(*------------------------------------------------------------------*)
(* ** Properties of star *)

Section Properties.

Hint Resolve hprop_empty_prove.

Lemma hprop_star_empty_l : forall H,
  hprop_empty \* H = H.
Proof using.
  intros. applys hprop_extens. intros h.
  iff (h1&h2&M1&M2&D&U) M.
  { forwards E: hprop_empty_inv M1. subst.
    rewrite~ heap_union_empty_l. }
  { exists~ heap_empty h. }
Qed.

Lemma hprop_star_comm : forall H1 H2,
   H1 \* H2 = H2 \* H1.
Proof using. 
  intros H1 H2. unfold hprop, hprop_star. extens. intros h.
  iff (h1&h2&M1&M2&D&U); rewrite~ heap_union_comm in U; exists* h2 h1.
Qed.

Lemma hprop_star_assoc : forall H1 H2 H3,
  (H1 \* H2) \* H3 = H1 \* (H2 \* H3).
Proof using.
  intros H1 H2 H3. unfold hprop, hprop_star. extens. intros h. split.
  { intros (h'&h3&(h1&h2&M3&M4&D'&U')&M2&D&U). subst h'.
    exists h1 (h2 \u h3). splits~.
    { exists h2 h3. splits*. }
    { subst. applys heap_eq. split. { state_eq. } { simpl; math. } } }
  { intros (h1&h'&M1&(h2&h3&M3&M4&D'&U')&D&U). subst h'.
    exists (h1 \u h2) h3. splits~.
    { exists h1 h2. splits*. } 
    { subst. applys heap_eq. split. { state_eq. } { simpl; math. } } }
Qed. 


(*------------------------------------------------------------------*)
(* ** Interaction of star with other operators *)

Lemma hprop_star_exists : forall A (J:A->hprop) H,
  (hprop_exists J) \* H = hprop_exists (fun x => (J x) \* H).
Proof using.
  intros. applys hprop_extens. intros h. iff M.
  { destruct M as (h1&h2&(x&M1)&M2&D&U). exists~ x h1 h2. }
  { destruct M as (x&(h1&h2&M1&M2&D&U)). exists h1 h2. splits~. exists~ x. }
Qed.

Lemma himpl_frame_l : forall H2 H1 H1',
  H1 ==> H1' -> 
  (H1 \* H2) ==> (H1' \* H2).
Proof using. introv W (h1&h2&?). exists* h1 h2. Qed.

End Properties.

End SepCreditsCore.




(********************************************************************)
(* * Properties of the logic *)

Module Export SepCreditsSetup := SepLogicSetup SepCreditsCore.


(*------------------------------------------------------------------*)
(* ** Singleton heap *)

Definition hprop_single (l:loc) (v:val) : hprop := 
  fun h => h^s = state_single l v /\ h^c = credits_zero /\ l <> null.

Notation "l '~~>' v" := (hprop_single l v)
  (at level 32, no associativity) : heap_scope.

Lemma hprop_star_single_same_loc_disjoint : forall (l:loc) (v1 v2:val),
  (l ~~> v1) \* (l ~~> v2) ==> \[False].
Proof using.
  intros. unfold hprop_single. 
  intros h ((m1&n1)&(m2&n2)&(E1&X1&N1)&(E2&X2&N2)&D&E). false.
  subst. applys* state_single_same_loc_disjoint l v1 v2.
  unfolds in D. rewrite <- E1. rewrite <- E2. auto.
Qed.

Global Opaque hprop_single.

(* ** Configure [hcancel] to make it aware of [hprop_single] *)

Ltac hcancel_hook H := 
  match H with
  | hprop_single _ _ => hcancel_try_same tt
  end.

Global Opaque hprop_single. 


(*------------------------------------------------------------------*)
(* ** Credits heap *)

Definition hprop_credits (n:nat) : hprop := 
  fun h => h^s = state_empty /\ h^c = n.

Notation "'\$' n" := (hprop_credits n) 
  (at level 40, format "\$ n") : heap_scope.

Lemma hprop_credits_inv : forall h n,
  (\$ n) h -> h = (state_empty,n).
Proof using. intros (m,n') n (M1&M2). simpls. subst*. Qed.

Global Opaque hprop_credits. 


(*------------------------------------------------------------------*)
(* ** Properties of credits *)

Section Credits.
Transparent hprop_credits hprop_empty hprop_pure hprop_star 
  heap_union heap_disjoint.

Lemma credits_zero_eq : \$ 0 = \[].
Proof using.
  unfold hprop_credits, hprop_empty, heap_empty. 
  applys prop_ext_1. intros [m n]; simpl. iff [M1 M2] M. 
  { subst~. }
  { inverts~ M. }
Qed.

Lemma credits_split_add : forall (n m : nat),
  \$ (n+m) = \$ n \* \$ m.
Proof using.
  intros c1 c2. unfold hprop_credits, hprop_star, heap_union, heap_disjoint. 
  applys prop_ext_1. intros [m n].
  iff [M1 M2] ([m1 n1]&[m2 n2]&(M1&E1)&(M2&E2)&M3&M4).
  { exists (state_empty:state,c1) (state_empty:state,c2). simpls. splits*. }
  { simpls. inverts M4. subst. split~. state_eq. }  
Qed.

Lemma credits_substract : forall (n m : nat), 
  (n >= m)%nat -> 
  \$ n ==> \$ m \* \$ (n-m).
Proof using.
  introv M. rewrite <- credits_split_add.
  math_rewrite (m + (n-m) = n)%nat. auto.
Qed.

End Credits.


(*------------------------------------------------------------------*)
(* ** Tactics for reductions *)

Ltac state_red_base tt ::=
  match goal with H: red _ _ ?t _ _ |- red _ _ ?t _ _ =>
    applys_eq H 2 4; try state_eq end.



(********************************************************************)
(* * Reasoning Rules *)

Implicit Types h : heap.
Implicit Types H : hprop.
Implicit Types Q : val -> hprop.


(*------------------------------------------------------------------*)
(* ** Definition of triples *)

Definition triple t H Q :=
  forall H' h, 
  (H \* H') h -> 
  exists n h' v, 
       red n (h^s) t (h'^s) v
    /\ (Q v \* \GC \* H') h'
    /\ (h^c = n + h'^c)%nat.


(*------------------------------------------------------------------*)
(* ** Definitions of [pay] *)

Definition pay_one H H' :=
  H ==> (\$ 1%nat) \* H'.


(*------------------------------------------------------------------*)
(* ** Structural rules *)

Lemma rule_extract_exists : forall t (A:Type) (J:A->hprop) Q,
  (forall x, triple t (J x) Q) ->
  triple t (hprop_exists J) Q.
Proof using.
  introv M. intros HF h N. rewrite hprop_star_exists in N.
  destruct N as (x&N). applys* M.
Qed.

Lemma rule_extract_prop : forall t (P:Prop) H Q,
  (P -> triple t H Q) ->
  triple t (\[P] \* H) Q.
Proof using.
  intros t. applys (rule_extract_prop_from_extract_exists (triple t)).
  applys rule_extract_exists.
Qed.

Lemma rule_consequence : forall t H' Q' H Q,
  H ==> H' ->
  triple t H' Q' ->
  Q' ===> Q ->
  triple t H Q.
Proof using. 
  introv MH M MQ. intros HF h N.
  forwards (n&h'&v&R&K&C): (rm M) HF h. { hhsimpl~. }
  exists n h' v. splits~. { hhsimpl~. }
Qed.

Lemma rule_frame : forall t H Q H',
  triple t H Q ->
  triple t (H \* H') (Q \*+ H').
Proof using. 
  introv M. intros HF h N. rewrite hprop_star_assoc in N.
  forwards (n&h'&v&R&K&C): (rm M) (H' \* HF) h. { hhsimpl~. }
  exists n h' v. splits~. { hhsimpl~. }
Qed.

Lemma rule_gc_post : forall t H Q,
  triple t H (Q \*+ \GC) ->
  triple t H Q.
Proof using.
  introv M. intros HF h N. forwards* (n&h'&v&R&K&C): (rm M) HF h.  
  exists n h' v. splits~. { rewrite <- hprop_gc_merge. hhsimpl. } 
Qed.

Lemma rule_gc_pre : forall t H Q,
  triple t H Q ->
  triple t (H \* \GC) Q.
Proof using.
  introv M. applys rule_gc_post. applys~ rule_frame.
Qed.


(*------------------------------------------------------------------*)
(* ** Term rules *)

Lemma rule_val : forall v H Q,
  H ==> Q v ->
  triple (trm_val v) H Q.
Proof using.
  introv M. intros HF h N. exists 0%nat h v. splits~.
  { applys red_val. }
  { hhsimpl. hchanges M. }
Qed.

Lemma rule_if : forall v t1 t2 H Q,
  triple (If v = val_int 0 then t2 else t1) H Q ->
  triple (trm_if v t1 t2) H Q.
Proof using.
  introv M. intros HF h N. forwards* (n&h'&v'&R&K&C): (rm M) HF h.
  exists n h' v'. splits~. { applys~ red_if. }
Qed.

Lemma rule_let : forall x t1 t2 H Q Q1,
  triple t1 H Q1 ->
  (forall (X:val), triple (subst_trm x X t2) (Q1 X) Q) ->
  triple (trm_let x t1 t2) H Q.
Proof using.
  introv M1 M2. intros HF h N.
  lets~ (n1&h1'&v1&R1&K1&C1): (rm M1) HF h.
  forwards* (n2&h2'&v2&R2&K2&C2): (rm M2) (\GC \* HF) h1'.
  exists (n1+n2)%nat h2' v2. splits~.
  { applys~ red_let R2. }
  { rewrite <- hprop_gc_merge. hhsimpl. } 
  { math. } 
Qed.

Lemma rule_let_val : forall x v1 t2 H Q,
  (forall (X:val), X = v1 -> triple (subst_trm x X t2) H Q) ->
  triple (trm_let x (trm_val v1) t2) H Q.
Proof using. 
  introv M. forwards~ M': (rm M).
  applys_eq~ (>> rule_let H (fun x => \[x = v1] \* H)) 2.
  { applys rule_val. hsimpl~. }
  { intros X. applys rule_extract_prop. intro_subst. applys M'. }
Qed.

Lemma rule_app : forall f x F V t1 H H' Q,
  F = (val_fix f x t1) ->
  pay_one H H' ->
  triple (subst_trm f F (subst_trm x V t1)) H' Q ->
  triple (trm_app F V) H Q.
Proof using.
  introv EF HP M. subst F. intros HF h N.
  unfolds pay_one.
  lets HP': himpl_frame_l HF (rm HP).
  lets N': (rm HP') (rm N). rew_heap in N'.
  destruct N' as (h1&h2&N1&N2&N3&N4).
  lets N1': hprop_credits_inv (rm N1). inverts N1'.
  lets (Na&Nb): heap_eq_forward (rm N4). simpls. subst. 
  lets~ (n&h'&v&R&K&C): (rm M) HF h2.
  exists (S n) h' v. splits~.
  { applys* red_app. state_red~. }
  { math. } 
Qed.

Lemma rule_let_fix : forall f x t1 t2 H Q,
  (forall (F:val), 
    (forall X H' H'' Q', 
      pay_one H' H'' ->
      triple (subst_trm f F (subst_trm x X t1)) H'' Q' ->
      triple (trm_app F X) H' Q') ->  
    triple (subst_trm f F t2) H Q) ->
  triple (trm_let f (trm_val (val_fix f x t1)) t2) H Q.
Proof using.
  introv M. applys rule_let_val. intros F EF. 
  applys (rm M). clears H Q. intros X H H' Q N. applys~ rule_app.
Qed.

Section RulesPrimitiveOps.
Transparent hprop_star hprop_single.

Lemma rule_ref : forall v,
  triple (prim_ref v) \[] (fun r => Hexists l, \[r = val_loc l] \* l ~~> v).
Proof using.
  intros. intros HF h N. rew_heap in N.
  forwards~ (l&Dl&Nl): (state_disjoint_new null (h^s) v).
  sets m1': (state_single l v).
  exists 0%nat ((m1' \+ h^s),h^c) (val_loc l). splits~.
  { applys~ red_ref. }
  { exists (m1',0%nat) h. split. 
    { exists l. unfold m1'. applys~ himpl_inst_prop (l ~~> v). hnfs~. }
    { splits~. hhsimpl~. } }
Qed.

Lemma rule_get : forall v l,
  triple (prim_get (val_loc l)) (l ~~> v) (fun x => \[x = v] \* (l ~~> v)).
Proof using.
  intros. intros HF h N. lets N': N.
  destruct N as (h1&h2&(N1a&N1b)&N2&N3&N4). 
  forwards (E1&E2): heap_eq_forward (rm N4). simpls.
  exists 0%nat h v. splits~.
  { applys red_get. rewrite E1. applys~ state_union_single_read. }
  { rew_heap. rewrite hprop_star_pure. split~. hhsimpl~. }
Qed.

Lemma rule_set : forall w l v, 
  triple (prim_set (val_pair (val_loc l) w)) (l ~~> v) (fun r => \[r = val_unit] \* l ~~> w).
Proof using.
  intros. intros HF h N. destruct N as (h1&h2&(N1a&N1b&N1c)&N2&N3&N4).
  forwards (E1&E2): heap_eq_forward (rm N4). simpls.
  sets m1': (state_single l w).
  exists 0%nat ((m1' \+ h2^s), h2^c) val_unit. splits~.
  { applys red_set. rewrite E1. unfold m1'. rewrite N1a.
    applys~ state_union_single_write. }
  { rew_heap. rewrite hprop_star_pure. split~.
    { exists (m1',0%nat) h2. splits~.
      { hnfs~. }
      { hhsimpl~. }
      { unfold m1'. unfolds heap_disjoint. rewrite N1a in N3.
        applys~ state_disjoint_single_set v. } } }
   { simpls. rewrite N1b in E2. unfolds credits_zero. math. }
Qed.

End RulesPrimitiveOps.


(********************************************************************)
(* * Bonus *)

(*------------------------------------------------------------------*)
(* ** Triples satisfy the [local] predicate *)

Lemma is_local_triple : forall t,
  is_local (triple t).
Proof using. 
  intros. applys prop_ext_2. intros H Q. iff M.
  { intros h Hh. forwards (h'&v&N1&N2): M \[] h.
    { hhsimpl. } 
    exists H \[] Q. splits~. { hhsimpl. } { hsimpl. } }
  { intros H' h Hh. lets (h1&h2&N1&N2&N3&N4): Hh. hnf in M.
    lets (H1&H2&Q1&R1&R2&R3): M N1.
    forwards (n&h'&v&S1&S2&S3): R2 (H2\*H') h.
    { subst h. rewrite <- hprop_star_assoc. exists~ h1 h2. }
    exists n h' v. splits~. rewrite <- hprop_gc_merge.
    applys himpl_forward S2.
    hchange (R3 v). rew_heap.
    rewrite (hprop_star_comm_assoc \GC H'). hsimpl. }  
Qed.


(*------------------------------------------------------------------*)
(* ** Alternative, slightly lower-level definition of triples *)

Definition triple' t H Q :=
  forall H' m c, 
  (H \* H') (m, c) -> 
  exists n m' c' v, 
       red n m t m' v
    /\ (Q v \* \GC \* H') (m', c')
    /\ (c = n + c')%nat.

Lemma triple_eq_triple' : triple = triple'.
Proof using.
  applys prop_ext_3. intros t H Q.
  unfold triple, triple'. iff M.
  { introv P1. forwards~ (n&h'&v&R1&R2&R3): M H' (m,c).
    exists n (h'^s) (h'^c) v. splits~. 
    applys_eq R2 1. applys~ heap_eq. } 
  { introv P1. forwards (n&m'&c'&v&R1&R2&R3): M H' (h^s) (h^c).
    { applys_eq P1 1. applys~ heap_eq. }
    exists n (m',c') v. splits~. } 
Qed.


(*------------------------------------------------------------------*)
(* ** Alternative, lower-level definition of triples *)

Definition triple'' t H Q :=
  forall m1 c1 m2, 
  state_disjoint m1 m2 -> 
  H (m1,c1) -> 
  exists n m1' m3' c1' v, 
       state_disjoint_3 m1' m3' m2 
    /\ red n (m1 \+ m2) t (m1' \+ m3' \+ m2) v
    /\ (Q v) (m1',c1')
    /\ (c1 >= n + c1')%nat.

Lemma triple_eq_triple'' : triple = triple''.
Proof using.
  hint hprop_gc_prove.
  applys prop_ext_3. intros t H Q.
  rewrite triple_eq_triple'.
  unfold triple', triple''. iff M.
  { introv D P1.
    forwards~ (n&m'&c'&v&R1&R2&R4): M (=(m2,0%nat)) (m1 \+ m2) c1.
    { exists (m1,c1) (m2,0%nat). splits~. applys heap_eq. simple~. } 
    rewrite <- hprop_star_assoc in R2.
    destruct R2 as ((m1''&c1'')&h2'&N0&N1&N2&N3). subst h2'.
    destruct N0 as ((m1'&c1')&(m3'&c3')&T0&T1&T2&T3). 
    exists n m1' m3' c1' v.
    lets (?&?): heap_eq_forward (rm T3).
    lets (?&?): heap_eq_forward (rm N3). simpls.
    splits~.
    { subst. rew_disjoint; simpls; rew_disjoint. destruct N2. 
     splits~. }
    { state_red. }
    { math. } }
  { introv (h1&h2&N1&N2&D&U).
    forwards~ (n&m1'&m3'&c1'&v&R1&R2&R3&R4): M (h1^s) (h1^c) (h2^s).
    { applys_eq N1 1. applys~ heap_eq. }
    lets (?&?): heap_eq_forward (rm U). simpls.
    exists n (m1' \+ m3' \+ h2^s) (c-n)%nat v. splits~.
    { state_red. } 
    { exists (m1',c1') (m3' \+ h2^s, (h2^c + h1^c - n - c1')%nat). splits~.
      { exists (m3',(h1^c - n - c1')%nat) h2. splits~. 
        { applys heap_eq. splits~. simpls. math. } } 
      { subst. rew_disjoint; simpls; rew_disjoint. autos*. }
      { applys heap_eq. splits~. simpls. subst. math. } }
    { math. }  }
Qed.

