(** 

This file formalizes the core constructs of "Separation Logic", 
as described in Arthur Charguéraud's lecture notes.

It consists of a functor, which may be used to construct, e.g.:
- a standard Separation Logic,
- a Separation Logic extended with credits,
- a Separation Logic extended with temporary read-only permissions.

The functor in this file assumes:
- a type for heaps
- a definition of \[P] (lifting of pure predicates) and of \* (star)
- a few properties of these operators

This file defines:
- heap entailement,
- derived heap operators: \[], (Hexists _,_), \GC, (x ~> R X)
- many useful lemmas,
- a couple of key tactics.

Author: Arthur Charguéraud.
License: MIT.

*)

Set Implicit Arguments.
Require Export LibCore.


(** Auxiliary tactic for obtaining a boolean answer
    to the question is E an evar?. 
    TODO: move to TLC *)

Ltac is_evar_as_bool E :=
  constr:(ltac:(first
    [ is_evar E; exact true
    | exact false ])).


(********************************************************************)
(** * Entailment *)

(*------------------------------------------------------------------*)
(* ** Definition of entailment *)

(** TLC defines [pred_le P Q] as [forall x, P x -> Q x]. *)

Notation "P ==> Q" := (pred_le P Q) 
  (at level 55, right associativity) : heap_scope.

(** TLC defines [rel_le P Q] as [forall x y, P x y -> Q x y]. *)

Notation "P ===> Q" := (rel_le P Q) 
  (at level 55, right associativity) : heap_scope.

Open Scope heap_scope.


(*------------------------------------------------------------------*)
(* ** Properties of entailment *)

Section HimplProp.
Variable A : Type.
Implicit Types H : A -> Prop.

Hint Unfold pred_le.

Lemma himpl_forward : forall H1 H2 (h:A),
  (H1 ==> H2) -> 
  (H1 h) -> 
  (H2 h).
Proof using. auto. Qed.

Lemma himpl_refl : forall H,
  H ==> H.
Proof using. intros h. auto. Qed.

Lemma himpl_trans : forall H1 H2 H3,
  (H1 ==> H2) -> 
  (H2 ==> H3) -> 
  (H1 ==> H3).
Proof using. introv M1 M2. intros h H1h. eauto. Qed.

Lemma himpl_antisym : forall H1 H2,
  (H1 ==> H2) -> 
  (H2 ==> H1) -> 
  (H1 = H2).
Proof using. introv M1 M2. applys prop_ext_1. intros h. iff*. Qed.

Lemma hprop_extens : forall H1 H2,
  (forall h, H1 h <-> H2 h) ->
  (H1 = H2).
Proof using.
  introv M. applys func_ext_1. intros h. applys* prop_ext.
Qed.

Lemma himpl_proj1 : forall H1 H2,
  H1 = H2 -> 
  H1 ==> H2.
Proof. intros. subst. applys~ himpl_refl. Qed.

Lemma himpl_proj2 : forall H1 H2,
  H1 = H2 -> 
  H2 ==> H1.
Proof. intros. subst. applys~ himpl_refl. Qed.

End HimplProp.

Implicit Arguments himpl_proj1 [A H1 H2].
Implicit Arguments himpl_proj2 [A H1 H2].
Implicit Arguments himpl_antisym [A H1 H2].

Hint Resolve himpl_refl rel_le_refl.


(********************************************************************)
(** * Abstract predicates for the Separation Logic *)

Module Type SepLogicCore.

(** Types *)

Parameter heap : Type.

Definition hprop := heap -> Prop.

(** Operators *)

Parameter hprop_empty : hprop.

Parameter hprop_star : hprop -> hprop -> hprop.

Definition hprop_exists A (J : A -> hprop) : hprop := 
  fun h => exists x, J x h.

Definition hprop_pure (P:Prop) : hprop := 
  hprop_exists (fun (p:P) => hprop_empty).

Definition hprop_gc := 
  hprop_exists (fun (H:hprop) => H).

(* Remark: [hprop_or] is not needed in practice.

  Definition hprop_or (H1 H2 : hprop) : hprop := 
    fun h => H1 h \/ H2 h.

*)

(** Notation *)

Notation "\[]" := (hprop_empty) 
  (at level 0) : heap_scope.
Notation "\[ P ]" := (hprop_pure P) 
  (at level 0, P at level 99) : heap_scope.
Notation "H1 '\*' H2" := (hprop_star H1 H2)
  (at level 41, right associativity) : heap_scope.
Notation "\GC" := (hprop_gc) : heap_scope. 
Open Scope heap_scope.

(** Properties of star *)

Parameter hprop_star_empty_l : forall H,
  \[] \* H = H.

Parameter hprop_star_comm : forall H1 H2,
   H1 \* H2 = H2 \* H1.

Parameter hprop_star_assoc : forall H1 H2 H3,
  (H1 \* H2) \* H3 = H1 \* (H2 \* H3).

(** Interaction of star with other operators *)

Parameter hprop_star_exists : forall A (J:A->hprop) H,
  (hprop_exists J) \* H = hprop_exists (fun x => (J x) \* H).

Parameter himpl_frame_l : forall H2 H1 H1',
  H1 ==> H1' -> 
  (H1 \* H2) ==> (H1' \* H2).

End SepLogicCore.


(********************************************************************)
(** * Functor *)

Module SepLogicSetup (SL : SepLogicCore).
Export SL.

Implicit Types h : heap.
Implicit Types H : hprop.


(*------------------------------------------------------------------*)
(* ** Representation predicates *)

Definition repr (A:Type) (S:A->hprop) (x:A) : hprop :=
  S x.

(** [x ~> Id X] holds when [x] is equal to [X] in the empty heap *)

Definition Id {A:Type} (X:A) (x:A) :=
  \[ X = x ].


(*------------------------------------------------------------------*)
(* ** Notation *)

Notation "x '~>' S" := (repr S x)
  (at level 33, no associativity) : heap_scope.

Notation "\[ L ]" := (hprop_pure L) 
  (at level 0, L at level 99) : heap_scope.

Notation "\[]" := (hprop_empty) 
  (at level 0) : heap_scope.

Notation "H1 '\*' H2" := (hprop_star H1 H2)
  (at level 41, right associativity) : heap_scope.

Notation "Q \*+ H" := (fun x => hprop_star (Q x) H)
  (at level 40) : heap_scope.

Notation "'Hexists' x1 , H" := (hprop_exists (fun x1 => H))
  (at level 39, x1 ident, H at level 50) : heap_scope.
Notation "'Hexists' x1 x2 , H" := (Hexists x1, Hexists x2, H)
  (at level 39, x1 ident, x2 ident, H at level 50) : heap_scope.
Notation "'Hexists' x1 x2 x3 , H" := (Hexists x1, Hexists x2, Hexists x3, H)
  (at level 39, x1 ident, x2 ident, x3 ident, H at level 50) : heap_scope.
Notation "'Hexists' x1 x2 x3 x4 , H" :=
  (Hexists x1, Hexists x2, Hexists x3, Hexists x4, H)
  (at level 39, x1 ident, x2 ident, x3 ident, x4 ident, H at level 50) : heap_scope.
Notation "'Hexists' x1 x2 x3 x4 x5 , H" :=
  (Hexists x1, Hexists x2, Hexists x3, Hexists x4, Hexists x5, H)
  (at level 39, x1 ident, x2 ident, x3 ident, x4 ident, x5 ident, H at level 50) : heap_scope.

Notation "'Hexists' x1 : T1 , H" := (hprop_exists (fun x1:T1 => H))
  (at level 39, x1 ident, H at level 50, only parsing) : heap_scope.
Notation "'Hexists' ( x1 : T1 ) , H" := (hprop_exists (fun x1:T1 => H))
  (at level 39, x1 ident, H at level 50, only parsing) : heap_scope.

Notation "\GC" := (hprop_gc) : heap_scope. 

Open Scope heap_scope.


(*------------------------------------------------------------------*)
(* ** Properties of types *)

Global Instance hprop_inhab : Inhab hprop.
Proof using. intros. apply (prove_Inhab hprop_empty). Qed.


(*------------------------------------------------------------------*)
(* ** Derived properties of [hprop] operators *)

Lemma hprop_star_empty_r : forall H,
  H \* \[] = H.
Proof using.
  applys neutral_r_from_comm_neutral_l. 
  applys~ hprop_star_comm. 
  applys~ hprop_star_empty_l.
Qed.

Lemma hprop_star_pure : forall (P:Prop) H h,
  (\[P] \* H) h = (P /\ H h).
Proof using.
  intros. extens. unfold hprop_pure.
  rewrite hprop_star_exists. 
  rewrite* hprop_star_empty_l.
  iff (p&M) (p&M). { split~. } { exists~ p. }
Qed.

Lemma hprop_pure_prove : forall (P:Prop) h,
  \[] h ->
  P -> 
  \[P] h.
Proof using. 
  introv M N. rewrite <- (hprop_star_empty_l \[P]).
  rewrite hprop_star_comm. rewrite~ hprop_star_pure.
Qed.

Lemma hprop_pure_inv : forall (P:Prop) h,
  \[P] h ->
  P /\ \[] h.
Proof using. 
  introv M. rewrite <- hprop_star_pure. 
  rewrite~ hprop_star_empty_r. 
Qed.

Lemma hprop_empty_def : 
  \[] = \[True].
Proof using. 
  applys prop_ext_1. intros h. iff M.
  { applys* hprop_pure_prove. }
  { forwards*: hprop_pure_inv M. }
Qed.

Lemma hprop_pure_def : forall (P:Prop),
  \[P] = (Hexists (p:P), \[]).
Proof using. auto. Qed.

Lemma hprop_gc_prove : forall h, 
  \GC h.
Proof using. intros. exists~ (=h). Qed.


(*------------------------------------------------------------------*)
(* ** Derived properties of [himpl] *)

Lemma himpl_frame_r : forall H1 H2 H2',
  H2 ==> H2' -> 
  (H1 \* H2) ==> (H1 \* H2').
Proof using.
  introv M. do 2 rewrite (@hprop_star_comm H1). applys~ himpl_frame_l.
Qed.

Lemma himpl_extract_prop : forall (P:Prop) H H',
  (P -> H ==> H') -> 
  (\[P] \* H) ==> H'.
Proof using.
  introv W Hh. rewrite hprop_star_pure in Hh. applys* W. 
Qed.

Lemma himpl_inst_prop : forall (P:Prop) H H',
  P -> 
  (H ==> H') ->
  H ==> (\[P] \* H').
Proof using.
  introv HP W. intros h Hh. rewrite~ hprop_star_pure.
Qed.

Lemma himpl_extract_exists : forall A H (J:A->hprop),
  (forall x, J x ==> H) -> (hprop_exists J) ==> H.
Proof using. introv W. intros h (x&Hh). applys* W. Qed.

Lemma himpl_inst_exists : forall A (x:A) H J,
  (H ==> J x) -> H ==> (hprop_exists J).
Proof using. introv W h. exists x. apply~ W. Qed.

Lemma himpl_remove_gc : forall H H',
  H ==> H' ->
  H ==> H' \* \GC.
Proof using.
  introv M. unfold hprop_gc. rewrite (hprop_star_comm H').
  rewrite hprop_star_exists.
  applys himpl_inst_exists \[]. rewrite~ hprop_star_empty_l.
Qed.

Lemma hprop_gc_merge : 
  \GC \* \GC = \GC.
Proof using.
  applys himpl_antisym. 
  { unfold hprop_gc. rewrite hprop_star_exists.
    applys himpl_extract_exists. intros H.
    rewrite hprop_star_comm.
    rewrite hprop_star_exists.
    applys himpl_extract_exists. intros H'.
    applys~ himpl_inst_exists (H' \* H). }
  { applys~ himpl_remove_gc. }
Qed.


(*------------------------------------------------------------------*)
(* ** Opaque operators *)

Global Opaque hprop_empty hprop_pure 
              hprop_star hprop_exists hprop_gc. 


(*------------------------------------------------------------------*)
(* ** Derived reasoning rules *)

Lemma rule_extract_prop_from_extract_exists :
  forall (T:Type) (F:hprop->(T->hprop)->Prop),
  (forall (A:Type) (J:A->hprop) (Q:T->hprop),
    (forall x, F (J x) Q) ->
    F (hprop_exists J) Q) ->
  (forall (P:Prop) H Q,
    (P -> F H Q) ->
    F (\[P] \* H) Q).
Proof using.
  introv M. introv N. rewrite hprop_pure_def.
  rewrite hprop_star_exists.
  applys M. rewrite~ hprop_star_empty_l.
Qed.

Implicit Arguments rule_extract_prop_from_extract_exists [T].




(*------------------------------------------------------------------*)
(*------------------------------------------------------------------*)
(*------------------------------------------------------------------*)
(* TODO: move the remaining of this file to a different file *)



(*------------------------------------------------------------------*)
(* ** [rew_heap] tactic to normalize expressions with [hprop_star] *)

Lemma star_post_empty : forall B (Q:B->hprop),
  Q \*+ \[] = Q.
Proof using. extens. intros. rewrite~ hprop_star_empty_r. Qed.

Hint Rewrite hprop_star_empty_l hprop_star_empty_r
             hprop_star_assoc star_post_empty : rew_heap.

Tactic Notation "rew_heap" :=
  autorewrite with rew_heap.
Tactic Notation "rew_heap" "in" "*" :=
  autorewrite with rew_heap in *.
Tactic Notation "rew_heap" "in" hyp(H) :=
  autorewrite with rew_heap in H.

Tactic Notation "rew_heap" "~" :=
  rew_heap; auto_tilde.
Tactic Notation "rew_heap" "~" "in" "*" :=
  rew_heap in *; auto_tilde.
Tactic Notation "rew_heap" "~" "in" hyp(H) :=
  rew_heap in H; auto_tilde.

Tactic Notation "rew_heap" "*" :=
  rew_heap; auto_star.
Tactic Notation "rew_heap" "*" "in" "*" :=
  rew_heap in *; auto_star.
Tactic Notation "rew_heap" "*" "in" hyp(H) :=
  rew_heap in H; auto_star.


(*------------------------------------------------------------------*)
(** Auxiliary lemmas useful for manual proofs *)

Lemma hprop_star_comm_assoc : forall H1 H2 H3,
  H1 \* H2 \* H3 = H2 \* H1 \* H3.
Proof using.
  intros. rewrite <- hprop_star_assoc. 
  rewrite (@hprop_star_comm H1 H2). rew_heap~.
Qed.



(*------------------------------------------------------------------*)
(** Auxiliary tactics used by many tactics *)

(* [xprecondition tt] returns the current pre-condition. *)

Ltac xprecondition tt :=
  match goal with |- ?R ?H ?Q => constr:(H) end.

(* [xpostcondition tt] returns the current post-condition. *)

Ltac xpostcondition tt :=
  match goal with |- ?E =>
  match get_fun_arg E with (_,?Q) => constr:(Q)
  end end.
  (* TODO: is this now equivalent to:
     match goal with |- ?J ?Q => constr:(Q) end. *)

(** [xpostcondition_is_evar tt] returns a boolean indicating
    whether the post-condition of the current goal is an evar. *)

Ltac xpostcondition_is_evar tt :=
  let Q := xpostcondition tt in
  is_evar_as_bool Q.


(*------------------------------------------------------------------*)
(** Auxiliary tactics used by [hpull] and [hsimpl] *)

Ltac remove_empty_heaps_from H :=
  match H with context[ ?H1 \* \[] ] =>
    rewrite (@hprop_star_empty_r H1) end.

Ltac remove_empty_heaps_left tt :=
  repeat match goal with |- ?H1 ==> _ => remove_empty_heaps_from H1 end.

Ltac remove_empty_heaps_right tt :=
  repeat match goal with |- _ ==> ?H2 => remove_empty_heaps_from H2 end.

Ltac on_formula_pre cont :=
  match goal with
  | |- _ ?H ?Q => cont H
  | |- _ _ ?H ?Q => cont H
  | |- _ _ _ ?H ?Q => cont H
  | |- _ _ _ _ ?H ?Q => cont H
  | |- _ _ _ _ _ ?H ?Q => cont H
  | |- _ _ _ _ _ _ ?H ?Q => cont H
  | |- _ _ _ _ _ _ _ ?H ?Q => cont H
  | |- _ _ _ _ _ _ _ _ ?H ?Q => cont H
  | |- _ _ _ _ _ _ _ _ _ ?H ?Q => cont H
  | |- _ _ _ _ _ _ _ _ _ _ ?H ?Q => cont H
  end.

Ltac on_formula_post cont :=
  match goal with
  | |- _ ?H ?Q => cont Q
  | |- _ _ ?H ?Q => cont Q
  | |- _ _ _ ?H ?Q => cont Q
  | |- _ _ _ _ ?H ?Q => cont Q
  | |- _ _ _ _ _ ?H ?Q => cont Q
  | |- _ _ _ _ _ _ ?H ?Q => cont Q
  | |- _ _ _ _ _ _ _ ?H ?Q => cont Q
  | |- _ _ _ _ _ _ _ _ ?H ?Q => cont Q
  | |- _ _ _ _ _ _ _ _ _ ?H ?Q => cont Q
  | |- _ _ _ _ _ _ _ _ _ _ ?H ?Q => cont Q
  end.

Ltac remove_empty_heaps_formula tt :=
  repeat (on_formula_pre ltac:(remove_empty_heaps_from)).




(*------------------------------------------------------------------*)
(** [hpull] tactic for extraction from [H1] on a goal [H1 ==> H2] *)

(** Lemmas *)

Lemma hpull_start : forall H H',
  \[] \* H ==> H' ->
  H ==> H'.
Proof using. intros. rew_heap in *. auto. Qed.

Lemma hpull_stop : forall H H',
  H ==> H' -> 
  H \* \[] ==> H'.
Proof using. intros. rew_heap in *. auto. Qed.

Lemma hpull_keep : forall H1 H2 H3 H',
  (H2 \* H1) \* H3 ==> H' ->
  H1 \* (H2 \* H3) ==> H'.
Proof using.
  intros. rewrite (hprop_star_comm H2) in H. rew_heap~ in *.
Qed.

Lemma hpull_starify : forall H1 H2 H',
  H1 \* (H2 \* \[]) ==> H' ->
  H1 \* H2 ==> H'.
Proof using. intros. rew_heap in *. auto. Qed.

Lemma hpull_assoc : forall H1 H2 H3 H4 H',
  H1 \* (H2 \* (H3 \* H4)) ==> H' ->
  H1 \* ((H2 \* H3) \* H4) ==> H'.
Proof using. intros. rew_heap in *. auto. Qed.

Lemma hpull_prop : forall H1 H2 H' (P:Prop),
  (P -> H1 \* H2 ==> H') -> 
  H1 \* (\[P] \* H2) ==> H'.
Proof using.
  intros. rewrite hprop_star_comm_assoc. applys~ himpl_extract_prop.
Qed.

Lemma hpull_empty : forall H1 H2 H',
  (H1 \* H2 ==> H') ->
  H1 \* (\[] \* H2) ==> H'. 
Proof using. intros. rew_heap~. Qed.

Lemma hpull_exists : forall A H1 H2 H' (J:A->hprop),
  (forall x, H1 \* J x \* H2 ==> H') -> 
  H1 \* (hprop_exists J \* H2) ==> H'.
Proof using.  
  intros. rewrite hprop_star_comm_assoc.
  rewrite hprop_star_exists.
  applys himpl_extract_exists. intros.
  rewrite~ <- hprop_star_comm_assoc.
Qed.

(** Tactics *)

Ltac hpull_prepare tt :=
  match goal with
  | |- @rel_le unit _ _ _ => let t := fresh "_tt" in intros t; destruct t
  | |- @rel_le _ _ _ _ => let r := fresh "r" in intros r
  | |- pred_le _ _ => idtac
  | _ => fail 1 "not a goal for hpull"
  end.

Ltac hpull_setup tt :=
  hpull_prepare tt;
  lets: ltac_mark;
  apply hpull_start.

Ltac hpull_cleanup tt :=
  apply hpull_stop;
  remove_empty_heaps_left tt;
  tryfalse;
  gen_until_mark.

Ltac hpull_step tt :=
  match goal with |- _ \* ?HN ==> _ => 
  match HN with 
  | ?H \* _ =>
     match H with
     | \[] => apply hpull_empty
     | \[_] => apply hpull_prop; intros
     | hprop_exists _ => apply hpull_exists; intros
     | _ \* _ => apply hpull_assoc
     | _ => apply hpull_keep
     end
  | \[] => fail 1
  | ?H => apply hpull_starify
  end end.

Ltac hpull_main tt :=
  hpull_setup tt;
  (repeat (hpull_step tt));
  hpull_cleanup tt.

Ltac hpull_post tt :=
  idtac. (* reflect_clean tt. *)

Ltac hpull_core :=
  hpull_main tt; hpull_post tt.

Tactic Notation "hpull" := 
  hpull_core.

(*-- Demo --*)

Lemma hpull_demo : forall H1 H2 H3 H,
   (H1 \* \[] \* (H2 \* Hexists (y:int), \[y = y]) \* H3) ==> H.
Proof using.
  intros. dup. 
  { hpull. admit. (* demo *) }
  { hpull_setup tt.
    hpull_step tt.
    hpull_step tt.
    hpull_step tt.
    hpull_step tt.
    hpull_step tt.
    hpull_step tt.
    hpull_step tt.
    hpull_step tt.
    try hpull_step tt.
    hpull_cleanup tt.
    admit. (* demo *)
Abort.


(*------------------------------------------------------------------*)
(* ** Tactic [hpull_check] tests whether it would be useful 
      to call [hpull] to extract things from the precondition.
      Applies to a goal of the form [H ==> H']. *)

(** Raises an error indicating the need to extract information *)

Ltac hpull_check_error tt :=
  fail 100 "need to first call hpull or xpull.".

(** [hpull_check_rec H] raises an error if the heap predicate [H]
    contains existentials or non-empty pure facts. *)

Ltac hpull_check_rec H :=
  let rec_after_simpl tt :=
    let H' := eval simpl in H in
     match H' with
     | H => hpull_check_error tt
     | _ => hpull_check_rec H'
     end
  in
  match H with
  | \[] => idtac
  | \[_ = _ :> unit] => idtac
  | \[_] => hpull_check_error tt 
  | hprop_exists _ => hpull_check_error tt
  | ?H1 \* ?H2 =>
     hpull_check_rec H1;
     hpull_check_rec H2
  | (fun _ => _) _ => rec_after_simpl tt
  | (let _ := _ in _) => rec_after_simpl tt
  | (let '(_,_) := _ in _) => rec_after_simpl tt
  | _ => idtac
  end.

(** [hpull_check tt] applies to a goal of the form (H ==> H')
    and raises an error if [H] contains extractible quantifiers
    or facts *)

Ltac hpull_check tt :=
  match goal with |- ?H ==> ?H' => hpull_check_rec H end.

(*-- Demo --*)

Lemma hpull_check_demo_1 : forall H1 H3 H,
  let H2 := Hexists (y:int), \[y = y] in
  (H1 \* H2 \* H3) ==> H.
Proof using. intros. hpull_check tt. (* --> accepts *) Abort.

Lemma hpull_check_demo_2 : forall H1 H2 H3 H,
  (H1 \* \[] \* (H2 \* Hexists (y:int), \[y = y]) \* H3) ==> H.
Proof using. intros. (* hpull_check tt. --> blocks *) Abort.


(*------------------------------------------------------------------*)
(* ** Tactic [hsimpl] to simplify a goal [H1 ==> H2] by cancelling
      equal items, and instantiating existentials from [H2]. *)

(* TODO: reimplement this tactic using arity-generic lemmas *)

(** Hints *)

Inductive Hcancel_hint : list Boxer -> Type :=
  | hcancel_hint : forall (L:list Boxer), Hcancel_hint L.

Ltac hcancel_hint_put L := 
  let H := fresh "Hint" in 
  generalize (hcancel_hint L); intros H.

Ltac hcancel_hint_next cont :=
  match goal with H: Hcancel_hint ((boxer ?x)::?L) |- _ =>
    clear H; hcancel_hint_put L; cont x end.

Ltac hcancel_hint_remove tt :=
  match goal with H: Hcancel_hint _ |- _ => clear H end.

(** Lemmas *)

Lemma hcancel_start : forall H' H1,
  H' ==> \[] \* H1 ->
  H' ==> H1.
Proof using. intros. rew_heap~ in *. Qed.

Lemma hcancel_stop : forall H' H1,
  H' ==> H1 ->
  H' ==> H1 \* \[].
Proof using. intros. rew_heap~ in *. Qed.

Lemma hcancel_keep : forall H' H1 H2 H3,
  H' ==> (H2 \* H1) \* H3 -> 
  H' ==> H1 \* (H2 \* H3).
Proof using. intros. rewrite (hprop_star_comm H2) in H. rew_heap~ in *. Qed.

Lemma hcancel_assoc : forall H' H1 H2 H3 H4,
  H' ==> H1 \* (H2 \* (H3 \* H4)) -> 
  H' ==> H1 \* ((H2 \* H3) \* H4).
Proof using. intros. rew_heap~ in *. Qed.

Lemma hcancel_starify : forall H' H1 H2,
  H' ==> H1 \* (H2 \* \[]) ->
  H' ==> H1 \* H2.
Proof using. intros. rew_heap~ in *. Qed.

Lemma hcancel_empty : forall H' H1 H2,
  H' ==> H1 \* H2 ->
  H' ==> H1 \* (\[] \* H2).
Proof using. intros. rew_heap~. Qed.

Lemma hcancel_prop : forall H' H1 H2 (P:Prop),
  H' ==> H1 \* H2 -> 
  P ->
  H' ==> H1 \* (\[P] \* H2).
Proof using.
  intros. rewrite hprop_star_comm_assoc. applys~ himpl_inst_prop.
Qed.

Lemma hcancel_exists : forall A (x:A) H' H1 H2 (J:A->hprop),
  H' ==> H1 \* J x \* H2 ->
  H' ==> H1 \* (hprop_exists J \* H2).
Proof using.
  intros. rewrite hprop_star_comm_assoc.
  rewrite hprop_star_exists.
  applys himpl_inst_exists x.
  rewrite~ hprop_star_comm_assoc.
Qed.

Lemma hcancel_id : forall A (x X : A) H' H1 H2,
  H' ==> H1 \* H2 -> x = X -> H' ==> H1 \* (x ~> Id X \* H2).
Proof using. intros. unfold Id. apply~ hcancel_prop. Qed.

Lemma hcancel_id_unify : forall A (x : A) H' H1 H2,
  H' ==> H1 \* H2 -> H' ==> H1 \* (x ~> Id x \* H2).
Proof using. intros. apply~ hcancel_id. Qed.

Lemma hcancel_gc : forall H, 
  H ==> \GC.
Proof using. 
  Transparent hprop_gc. intros. unfold hprop_gc. introv M. exists~ H.
Qed.

Lemma hcancel_cancel_1 : forall H HA HR HT,
  HT ==> HA \* HR ->
  H \* HT ==> HA \* (H \* HR).
Proof using. intros. rewrite hprop_star_comm_assoc. applys~ himpl_frame_r. Qed.

Lemma hcancel_cancel_2 : forall H HA HR H1 HT,
  H1 \* HT ==> HA \* HR -> 
  H1 \* H \* HT ==> HA \* (H \* HR).
Proof using. intros. rewrite (hprop_star_comm_assoc H1). apply~ hcancel_cancel_1. Qed.

Lemma hcancel_cancel_3 : forall H HA HR H1 H2 HT,
  H1 \* H2 \* HT ==> HA \* HR -> H1 \* H2 \* H \* HT ==> HA \* (H \* HR).
Proof using. intros. rewrite (hprop_star_comm_assoc H2). apply~ hcancel_cancel_2. Qed.

Lemma hcancel_cancel_4 : forall H HA HR H1 H2 H3 HT,
  H1 \* H2 \* H3 \* HT ==> HA \* HR ->
  H1 \* H2 \* H3 \* H \* HT ==> HA \* (H \* HR).
(*Proof using. intros. rewrite (hprop_star_comm_assoc H3). apply~ hcancel_cancel_3. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_cancel_5 : forall H HA HR H1 H2 H3 H4 HT,
  H1 \* H2 \* H3 \* H4 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H4 \* H \* HT ==> HA \* (H \* HR).
(*Proof using. intros. rewrite (hprop_star_comm_assoc H4). apply~ hcancel_cancel_4. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_cancel_6 : forall H HA HR H1 H2 H3 H4 H5 HT,
  H1 \* H2 \* H3 \* H4 \* H5 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H \* HT ==> HA \* (H \* HR).
(*Proof using. intros. rewrite (hprop_star_comm_assoc H5). apply~ hcancel_cancel_5. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_cancel_7 : forall H HA HR H1 H2 H3 H4 H5 H6 HT,
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H \* HT ==> HA \* (H \* HR).
(*Proof using. intros. rewrite (hprop_star_comm_assoc H6). apply~ hcancel_cancel_6. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_cancel_8 : forall H HA HR H1 H2 H3 H4 H5 H6 H7 HT,
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H \* HT ==> HA \* (H \* HR).
(*Proof using. intros. rewrite (star_comm_assoc H7). apply~ hcancel_cancel_7. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_cancel_9 : forall H HA HR H1 H2 H3 H4 H5 H6 H7 H8 HT,
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H \* HT ==> HA \* (H \* HR).
(*Proof using. intros. rewrite (hprop_star_comm_assoc H8). apply~ hcancel_cancel_8. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_cancel_10 : forall H HA HR H1 H2 H3 H4 H5 H6 H7 H8 H9 HT,
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 \* H \* HT ==> HA \* (H \* HR).
(*Proof using. intros. rewrite (hprop_star_comm_assoc H9). apply~ hcancel_cancel_9. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_cancel_eq_1 : forall H H' HA HR HT,
  H = H' -> HT ==> HA \* HR -> H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_1. Qed.

Lemma hcancel_cancel_eq_2 : forall H H' HA HR H1 HT,
  H = H' -> H1 \* HT ==> HA \* HR -> 
  H1 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_2. Qed.

Lemma hcancel_cancel_eq_3 : forall H H' HA HR H1 H2 HT,
  H = H' -> H1 \* H2 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_3. Qed.

Lemma hcancel_cancel_eq_4 : forall H H' HA HR H1 H2 H3 HT,
  H = H' -> H1 \* H2 \* H3 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_4. Qed.

Lemma hcancel_cancel_eq_5 : forall H H' HA HR H1 H2 H3 H4 HT,
  H = H' -> H1 \* H2 \* H3 \* H4 \* HT ==> HA \* HR -> 
  H1 \* H2 \* H3 \* H4 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_5. Qed.

Lemma hcancel_cancel_eq_6 : forall H H' HA HR H1 H2 H3 H4 H5 HT,
  H = H' -> H1 \* H2 \* H3 \* H4 \* H5 \* HT ==> HA \* HR ->
  H1 \* H2 \* H3 \* H4 \* H5 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_6. Qed.

Lemma hcancel_cancel_eq_7 : forall H H' HA HR H1 H2 H3 H4 H5 H6 HT,
  H = H' -> H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* HT ==> HA \* HR ->
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_7. Qed.

Lemma hcancel_cancel_eq_8 : forall H H' HA HR H1 H2 H3 H4 H5 H6 H7 HT,
  H = H' -> H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* HT ==> HA \* HR ->
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_8. Qed.

Lemma hcancel_cancel_eq_9 : forall H H' HA HR H1 H2 H3 H4 H5 H6 H7 H8 HT,
  H = H' -> H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* HT ==> HA \* HR ->
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_9. Qed.

Lemma hcancel_cancel_eq_10 : forall H H' HA HR H1 H2 H3 H4 H5 H6 H7 H8 H9 HT,
  H = H' -> H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 \* HT ==> HA \* HR ->
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 \* H \* HT ==> HA \* (H' \* HR).
Proof using. intros. subst. apply~ hcancel_cancel_10. Qed.

Lemma hcancel_start_1 : forall H1 H', 
  H1 \* \[] ==> H' -> H1 ==> H'.
Proof using. intros. rew_heap in H. auto. Qed.

Lemma hcancel_start_2 : forall H1 H2 H', 
  H1 \* H2 \* \[] ==> H' -> 
  H1 \* H2 ==> H'.
Proof using. intros. rew_heap in H. auto. Qed.

Lemma hcancel_start_3 : forall H1 H2 H3 H', 
  H1 \* H2 \* H3 \* \[] ==> H' -> 
  H1 \* H2 \* H3 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed.*)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_start_4 : forall H1 H2 H3 H4 H', 
  H1 \* H2 \* H3 \* H4 \* \[] ==> H' -> 
  H1 \* H2 \* H3 \* H4 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed. *)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_start_5 : forall H1 H2 H3 H4 H5 H', 
  H1 \* H2 \* H3 \* H4 \* H5 \* \[] ==> H' -> 
  H1 \* H2 \* H3 \* H4 \* H5 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed. *)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_start_6 : forall H1 H2 H3 H4 H5 H6 H', 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* \[] ==> H' -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed. *)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_start_7 : forall H1 H2 H3 H4 H5 H6 H7 H', 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* \[] ==> H' -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed. *)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_start_8 : forall H1 H2 H3 H4 H5 H6 H7 H8 H', 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* \[] ==> H' -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed. *)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_start_9 : forall H1 H2 H3 H4 H5 H6 H7 H8 H9 H', 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 \* \[] ==> H' -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed. *)
Admitted. (* commented out for faster compilation *)

Lemma hcancel_start_10 : forall H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H', 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 \* H10 \* \[] ==> H' -> 
  H1 \* H2 \* H3 \* H4 \* H5 \* H6 \* H7 \* H8 \* H9 \* H10 ==> H'.
(* Proof using. intros. rew_heap in H. auto. Qed. *)
Admitted. (* commented out for faster compilation *)

(** Tactics *)


Ltac hcancel_left_empty tt :=
  match goal with |- ?HL ==> _ =>
  match HL with
  | \[] => idtac
  | _ \* \[] => idtac
  | _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* \[] => idtac
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* ?H => apply hcancel_start_10
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* ?H => apply hcancel_start_9
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* ?H => apply hcancel_start_8
  | _ \* _ \* _ \* _ \* _ \* _ \* ?H => apply hcancel_start_7
  | _ \* _ \* _ \* _ \* _ \* ?H => apply hcancel_start_6
  | _ \* _ \* _ \* _ \* ?H => apply hcancel_start_5
  | _ \* _ \* _ \* ?H => apply hcancel_start_4
  | _ \* _ \* ?H => apply hcancel_start_3
  | _ \* ?H => apply hcancel_start_2
  | ?H => apply hcancel_start_1
  end end.


Ltac check_arg_true v :=
  match v with
  | true => idtac
  | false => fail 1 
  end.

Ltac hcancel_prepare tt :=
  match goal with
  (* LATER | |- @rel_le unit _ ?Q ?Q' => let t := fresh "_tt" in intros t; destruct t *)
  | |- @rel_le _ _ _ _ => let r := fresh "r" in intros r
  | |- pred_le _ ?H' => is_evar H'; apply himpl_refl
  | |- pred_le _ _ => idtac
  end.

Ltac hcancel_setup tt := 
  hcancel_prepare tt;
  hcancel_left_empty tt;
  apply hcancel_start.

Ltac hcancel_cleanup tt :=
  try apply hcancel_stop;
  try apply hcancel_stop;
  try apply pred_le_refl;
  try hcancel_hint_remove tt;
  try remove_empty_heaps_right tt;
  try remove_empty_heaps_left tt;
  try apply hcancel_gc.

Ltac hcancel_try_same tt :=
  first 
  [ apply hcancel_cancel_1
  | apply hcancel_cancel_2
  | apply hcancel_cancel_3
  | apply hcancel_cancel_4
  | apply hcancel_cancel_5
  | apply hcancel_cancel_6 
  | apply hcancel_cancel_7 
  | apply hcancel_cancel_8 
  | apply hcancel_cancel_9 
  | apply hcancel_cancel_10 
  ].

Ltac hcancel_find_same H HL :=
  match HL with
  | H \* _ => apply hcancel_cancel_1
  | _ \* H \* _ => apply hcancel_cancel_2
  | _ \* _ \* H \* _ => apply hcancel_cancel_3
  | _ \* _ \* _ \* H \* _ => apply hcancel_cancel_4
  | _ \* _ \* _ \* _ \* H \* _ => apply hcancel_cancel_5
  | _ \* _ \* _ \* _ \* _ \* H \* _ => apply hcancel_cancel_6
  | _ \* _ \* _ \* _ \* _ \* _ \* H \* _ => apply hcancel_cancel_7
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* H \* _ => apply hcancel_cancel_8
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* H \* _ => apply hcancel_cancel_9
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* H \* _ => apply hcancel_cancel_10
  end.

Ltac hcancel_find_repr l HL cont :=
  match HL with
  | repr _ l \* _ => apply hcancel_cancel_eq_1
  | _ \* repr _ l \* _ => apply hcancel_cancel_eq_2
  | _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_3
  | _ \* _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_4
  | _ \* _ \* _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_5
  | _ \* _ \* _ \* _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_6
  | _ \* _ \* _ \* _ \* _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_7
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_8
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_9
  | _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* _ \* repr _ l \* _ => apply hcancel_cancel_eq_10
  end; [ cont tt | ].


Ltac hcancel_extract_exists tt :=
  first [ 
    hcancel_hint_next ltac:(fun x =>
      match x with
      | __ => eapply hcancel_exists
      | _ => apply (@hcancel_exists _ x)
      end)
  | eapply hcancel_exists ].

Ltac hcancel_find_repr_post tt :=
  try solve 
   [ reflexivity
   | fequal; fequal; first [ eassumption | symmetry; eassumption ] ].

(** Maintain the goal in the form
     H1 \* ... \* HN \* [] ==> HA \* HR
   where HA is initially empty and accumulates elements not simplifiable
   and HR contains the values that are to be cancelled out;
   the last item of HR is always a [].
   As long as HR is of the form H \* H', we try to match H with one of the Hi.
  *)

Ltac hcancel_hook H := fail.

Ltac check_noevar2 M := (* todo: merge *)
  first [ has_evar M; fail 1 | idtac ].

Ltac hcancel_step tt :=
  match goal with |- ?HL ==> ?HA \* ?HN =>
  match HN with
  | ?H \* _ =>
    match H with
    | ?H => hcancel_hook H
    | \[] => apply hcancel_empty
    | \[_] => apply hcancel_prop
    | hprop_exists _ => hcancel_extract_exists tt
    | _ \* _ => apply hcancel_assoc
    | ?H => 
       first [ is_evar H; fail 1 | idtac ];
       hcancel_find_same H HL (* may fail *)
    | ?x ~> _ => hcancel_find_repr x HL ltac:(hcancel_find_repr_post) (* may fail *)
    | ?x ~> Id _ => check_noevar2 x; apply hcancel_id (* may fail *)
    | ?x ~> ?T _ => check_noevar2 x;
                    let M := fresh in assert (M: T = Id); [ reflexivity | clear M ];
                    apply hcancel_id; [ | reflexivity ]
                    (* may fail *)
    | ?x ~> ?T ?X => check_noevar2 x; is_evar T; is_evar X; apply hcancel_id_unify
    | _ => apply hcancel_keep
    end
  | \[] => fail 1
  | _ => apply hcancel_starify
  end end.

(* TODO: factorize the logging version of the code with the normal code *)
Ltac hcancel_step_debug tt :=
  match goal with |- ?HL ==> ?HA \* ?HN =>
  idtac HN;
  match HN with
  | ?H \* _ =>
    match H with
    | ?H => hcancel_hook H; idtac "hook"
    | \[] => apply hcancel_empty
    | \[_] => apply hcancel_prop
    | hprop_exists _ => hcancel_extract_exists tt
    | _ \* _ => idtac "sep"; apply hcancel_assoc
    | ?H => idtac "find"; 
        first [ has_evar H; idtac "has evar"; fail 1 | idtac "has no evar" ];
         hcancel_find_same H HL (* may fail *)
    | ?x ~> _ => idtac "repr"; hcancel_find_repr x HL ltac:(hcancel_find_repr_post) (* may fail *)
    | ?x ~> Id _ => idtac "id";  check_noevar x; apply hcancel_id (* todo: why is this requiring a fail 2 ? *)
    | ?x ~> _ _ => idtac "some";  check_noevar2 x; apply hcancel_id_unify
    | ?X => idtac "keep"; apply hcancel_keep
    end
  | \[] => fail 1
  | _ => idtac "starify"; apply hcancel_starify
  end end.

Ltac hcancel_main tt :=
  hcancel_setup tt;
  (repeat (hcancel_step tt));
  hcancel_cleanup tt.


Tactic Notation "hcancel" := 
  hcancel_main false.

Tactic Notation "hcancel_core" constr(L) :=
  match type of L with 
  | list Boxer => hcancel_hint_put L
  | _ => hcancel_hint_put (boxer L :: nil)
  end; 
  hcancel.

Ltac hsimpl_core tt := 
  hpull; intros; hcancel.

Tactic Notation "hsimpl" := hsimpl_core tt.
Tactic Notation "hsimpl" "~" := hsimpl; auto_tilde.
Tactic Notation "hsimpl" "*" := hsimpl; auto_star.
Tactic Notation "hsimpl" constr(L) :=
  match type of L with 
  | list Boxer => hcancel_hint_put L
  | _ => hcancel_hint_put (boxer L :: nil)
  end; hsimpl.
Tactic Notation "hsimpl" constr(X1) constr(X2) :=
  hsimpl (>> X1 X2).
Tactic Notation "hsimpl" constr(X1) constr(X2) constr(X3) :=
  hsimpl (>> X1 X2 X3).

Tactic Notation "hsimpl" "~" constr(L) :=
  hsimpl L; auto_tilde.
Tactic Notation "hsimpl" "~" constr(X1) constr(X2) :=
  hsimpl X1 X2; auto_tilde.
Tactic Notation "hsimpl" "~" constr(X1) constr(X2) constr(X3) :=
  hsimpl X1 X2 X3; auto_tilde.

Tactic Notation "hsimpl" "*" constr(L) :=
  hsimpl L; auto_star.
Tactic Notation "hsimpl" "*" constr(X1) constr(X2) :=
  hsimpl X1 X2; auto_star.
Tactic Notation "hsimpl" "*" constr(X1) constr(X2) constr(X3) :=
  hsimpl X1 X2 X3; auto_star.


(*-- Demo --*)

Lemma hsimpl_demo_1 : forall H1 H2 H3 H4 H5,
  H1 \* H2 \* H3 \* H4 ==> H4 \* H3 \* H5 \* H2.
Proof using.
  intros. dup.
  { hsimpl. admit. (* demo *) }
  { hcancel_setup tt.
    hcancel_step tt.
    hcancel_step tt.
    hcancel_step tt.
    hcancel_step tt.
    hcancel_step tt.
    try hcancel_step tt.
    hcancel_cleanup tt.
    admit. (* demo *) }
Abort.

Lemma hsimpl_demo_2 : forall H1 H2 H3 H4 H5,
  H1 \* H2 \* H3 \* H4 \* H5 ==> H3 \* H1 \* H2 \* \GC.
Proof using. intros. hsimpl. Abort.

Lemma hsimpl_demo_3 : forall H1 H2,
  (forall H, H1 \* H2 ==> H -> True) -> True.
Proof using. intros. eapply H. hsimpl. Abort.

Lemma demo_hsimpl_hints : exists n, n = 3.
Proof using.
  hcancel_hint_put (>> 3 true).
  hcancel_hint_next ltac:(fun x => exists x).
  hcancel_hint_remove tt.
Abort.


(*------------------------------------------------------------------*)
(* ** Tactic [hhsimpl] to prove [H h] from [H' h] *)

Ltac hhsimpl_core :=
  match goal with N: ?H ?h |- _ ?h => 
    match type of H with hprop =>
    applys himpl_forward N; clear N; hsimpl 
  end end.

Tactic Notation "hhsimpl" := hhsimpl_core.
Tactic Notation "hhsimpl" "~" := hhsimpl; auto_tilde.
Tactic Notation "hhsimpl" "*" := hhsimpl; auto_star.


(*------------------------------------------------------------------*)
(* ** Tactic [hchange] *)

Lemma hchange_lemma : forall H1 H1' H H' H2,
  (H1 ==> H1') -> 
  (H ==> H1 \* H2) -> 
  (H1' \* H2 ==> H') -> 
  (H ==> H').
Proof using.
  intros. applys* (@pred_le_trans heap) (H1 \* H2).
  applys* (@pred_le_trans heap) (H1' \* H2). hsimpl~.
Qed.

Ltac hchange_apply L cont1 cont2 :=
  eapply hchange_lemma;
    [ applys L | cont1 tt | cont2 tt ].

Ltac hchange_forwards L modif cont1 cont2 :=
  forwards_nounfold_then L ltac:(fun K =>
  match modif with
  | __ =>
     match type of K with
     | _ = _ => hchange_apply (@himpl_proj1 _ _ _ K) cont1 cont2
     | _ => hchange_apply K cont1 cont2
     end
  | _ => hchange_apply (@modif _ _ _ K) cont1 cont2
  end).

Ltac hcancel_cont tt :=
  instantiate; hcancel.

Ltac hsimpl_cont tt :=
  instantiate; hsimpl.

Ltac hchange_core E modif cont1 cont2 :=
  hpull; intros;
  match E with
  (*  | ?H ==> ?H' => hchange_with_core H H' --todo*)
  | _ => hchange_forwards E modif ltac:(cont1) ltac:(cont2)
  end.

Ltac hchange_debug_base E modif :=
  hchange_forwards E modif ltac:(idcont) ltac:(idcont).

Tactic Notation "hchange_debug" constr(E) :=
  hchange_debug_base E __.
Tactic Notation "hchange_debug" "->" constr(E) :=
  hchange_debug_base E himpl_proj1.
Tactic Notation "hchange_debug" "<-" constr(E) :=
  hchange_debug_base himpl_proj2.

Ltac hchange_base E modif :=
  hchange_core E modif ltac:(hcancel_cont) ltac:(idcont).

Tactic Notation "hchange" constr(E) :=
  hchange_base E __.
Tactic Notation "hchange" "->" constr(E) :=
  hchange_base E himpl_proj1.
Tactic Notation "hchange" "<-" constr(E) :=
  hchange_base E himpl_proj2.

Tactic Notation "hchange" "~" constr(E) :=
  hchange E; auto_tilde.
Tactic Notation "hchange" "*" constr(E) :=
  hchange E; auto_star.

Ltac hchanges_base E modif :=
  hchange_core E modif ltac:(hcancel_cont) ltac:(hsimpl_cont).

Tactic Notation "hchanges" constr(E) :=
  hchanges_base E __.
Tactic Notation "hchanges" "->" constr(E) :=
  hchanges_base E himpl_proj1.
Tactic Notation "hchange" "<-" constr(E) :=
  hchanges_base E himpl_proj2.

Tactic Notation "hchanges" "~" constr(E) :=
  hchanges E; auto_tilde.
Tactic Notation "hchanges" "*" constr(E) :=
  hchanges E; auto_star.

Tactic Notation "hchange" constr(E1) constr(E2) :=
  hchange E1; hchange E2.
Tactic Notation "hchange" constr(E1) constr(E2) constr(E3) :=
  hchange E1; hchange E2 E3.


(*------------------------------------------------------------------*)
(* ** Tactic [xunfold] to unfold heap predicates *)

Tactic Notation "ltac_set" "(" ident(X) ":=" constr(E) ")" "at" constr(K) :=
  match nat_from_number K with
  | 1%nat => set (X := E) at 1
  | 2%nat => set (X := E) at 2
  | 3%nat => set (X := E) at 3
  | 4%nat => set (X := E) at 4
  | 5%nat => set (X := E) at 5
  | 6%nat => set (X := E) at 6
  | 7%nat => set (X := E) at 7
  | 8%nat => set (X := E) at 8
  | 9%nat => set (X := E) at 9
  | 10%nat => set (X := E) at 10
  | 11%nat => set (X := E) at 11
  | 12%nat => set (X := E) at 12
  | 13%nat => set (X := E) at 13
  | _ => fail "ltac_set: arity not supported"
  end.

(** [xunfold at n] unfold the definition of the arrow [~>]
    at the occurence [n] in the goal. *)

Definition repr' (A:Type) (S:A->hprop) (x:A) : hprop := S x.

Tactic Notation "xunfold" "at" constr(n) :=
  let h := fresh "temp" in
  ltac_set (h := repr) at n;
  change h with repr';
  unfold repr';
  clear h.

(** [xunfold_clean] simplifies instances of
   [p ~> (fun _ => _)] by unfolding the arrow,
   but only when the body does not captures
   locally-bound variables.
   TODO: deprecated *)

Tactic Notation "xunfold_clean" :=
  try match goal with |- context C [?p ~> ?E] =>
   match E with (fun _ => _) =>
     let E' := eval cbv beta in (E p) in
     let G' := context C [E'] in
     let G := match goal with |- ?G => G end in
     change G with G' end end.


(** [xunfold E] unfolds all occurences of the representation
    predicate [E].
    Limitation: won't work if E has more than 12 arguments.

    Implementation: converts all occurences of repr to repr',
    then unfolds these occurences one by one, and considers
    them for unfolding. *)

Tactic Notation "xunfold" constr(E) :=
  change repr with repr';
  let h := fresh "temp" in
  set (h := repr');
  repeat (
    unfold h at 1;
    let ok := match goal with
      | |- context [ repr' (E) _ ] => constr:(true)
      | |- context [ repr' (E _) _ ] => constr:(true)
      | |- context [ repr' (E _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
      | |- context [ repr' (E _ _ _ _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
      | _ => constr:(false)
      end in
    match ok with
    | true => unfold repr'
    | false => change repr' with repr
    end);
  clear h;
  simpl E.
    (* TODO: was "unfold E", but this fails for recursive definitions. *)

Tactic Notation "xunfold" "~" constr(E) := xunfold E; auto_tilde.
Tactic Notation "xunfold" "*" constr(E) := xunfold E; auto_star.


(** [xunfold E] unfolds a specific occurence of the representation
    predicate [E]. TODO: still needs to unfold [E] by hand. *)

Tactic Notation "xunfold" constr(E) "at" constr(n) :=
  let n := nat_from_number n in
  change repr with repr';
  let h := fresh "temp" in
  set (h := repr');
  let E' := fresh "tempR" in
  set (E' := E);
  let rec aux n :=
    try (
      unfold h at 1;
      let ok := match goal with
        | |- context [ repr' (E') _ ] => constr:(true)
        | |- context [ repr' (E' _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
        | |- context [ repr' (E' _ _ _ _ _ _ _ _ _ _ _ _) _ ] => constr:(true)
        | _ => constr:(false)
        end in
      match ok with
      | true =>
         match n with
         | (S O)%nat =>
            unfold repr'
         | (S ?n')%nat => change repr' with repr; aux n'
         end
      | false => change repr' with repr; aux n
      end)
     in
  aux n;
  unfold h;
  clear h;
  change repr' with repr;
  unfold E';
  clear E'.



(*------------------------------------------------------------------*)
(* ** Definition and properties of [local] *)

(** Type of characteristic formulae on values of type B *)

Notation "'~~' B" := (hprop->(B->hprop)->Prop)
  (at level 8, only parsing) : type_scope.

(** "Local" = Frame rule + consequence rule + garbage collection *)

Definition local B (F:~~B) : ~~B :=
  fun (H:hprop) (Q:B->hprop) =>
    forall h, H h -> exists H1 H2 Q1,
       (H1 \* H2) h
    /\ F H1 Q1
    /\ Q1 \*+ H2 ===> Q \*+ \GC.

(** Characterization of "local" judgments *)

Definition is_local B (F:~~B) :=
  F = local F.

(** [is_local_pred S] asserts that [is_local (S x)] holds for any [x].
    It is useful for describing loop invariants. *)

Definition is_local_pred A B (S:A->~~B) :=
  forall x, is_local (S x).


(*------------------------------------------------------------------*)
(* ** Properties of [local] *)

(** The [local] operator can be freely erased from a conclusion *)

Lemma local_erase : forall B (F:~~B),
  forall H Q, F H Q -> local F H Q.
Proof using.
  intros. exists H \[] Q. splits.
  rew_heap~. auto. hsimpl.
Qed.

(** A introduction rule to establish [is_local] *)

Lemma is_local_prove : forall B (F:~~B),
  (forall H Q, F H Q <-> local F H Q) -> is_local F.
Proof using.
  intros. unfold is_local. apply func_ext_2.
  intros. applys prop_ext. applys H.
Qed.

(** Weaken and frame and gc property [local] *)

Lemma local_frame_gc : forall B (F:~~B) H H1 H2 Q1 Q,
  is_local F ->
  F H1 Q1 ->
  H ==> H1 \* H2 ->
  Q1 \*+ H2 ===> Q \*+ \GC ->
  F H Q.
Proof using.
  introv L M WH WQ. rewrite L. introv Ph.
  exists H1 H2 Q1. splits; rew_heap~.
Qed.

(** Weaken and frame properties from [local] *)

Lemma local_frame : forall B H1 H2 Q1 (F:~~B) H Q,
  is_local F ->
  F H1 Q1 ->
  H ==> H1 \* H2 ->
  Q1 \*+ H2 ===> Q ->
  F H Q.
Proof using.
  introv L M WH WQ. rewrite L. introv Ph.
  exists H1 H2 Q1. splits; rew_heap~.
  hchange WQ. hsimpl.
Qed.

(** Weakening on pre and post from [local] *)

Lemma local_weaken : forall B H' Q' (F:~~B) H Q,
  is_local F ->
  F H' Q' ->
  H ==> H' ->
  Q' ===> Q ->
  F H Q.
Proof using.
  intros. eapply local_frame with (H2 := \[]); eauto; rew_heap~.
Qed.

(** Garbage collection on precondition from [local] *)

Lemma local_gc_pre : forall H' B (F:~~B) H Q,
  is_local F ->
  H ==> H' \* \GC ->
  F H' Q ->
  F H Q.
Proof using.
  introv LF H1 H2. applys~ local_frame_gc H2 H1.
Qed.

(** Garbage collection on post-condition from [local] *)

Lemma local_gc_post : forall B Q' (F:~~B) H Q,
  is_local F ->
  F H Q' ->
  Q' ===> Q \*+ \GC ->
  F H Q.
Proof using.
  introv L M W. rewrite L. introv Ph.
  exists H \[] Q'. splits; rew_heap~.
Qed.

(** Variant of the above, useful for tactics to specify
    the garbage collected part *)

Lemma local_gc_pre_on : forall HG H' B (F:~~B) H Q,
  is_local F ->
  H ==> HG \* H' ->
  F H' Q ->
  F H Q.
Proof using.
  introv L M W. rewrite L. introv Ph.
  exists H' HG Q. splits.
  rewrite hprop_star_comm. apply~ M.
  auto.
  hsimpl.
Qed.

(** Weakening on pre and post with gc -postfrom [local] *)

Lemma local_weaken_gc : forall B H' Q' (F:~~B) H Q,
  is_local F ->
  F H' Q' ->
  H ==> H' ->
  Q' ===> Q \*+ \GC ->
  F H Q.
Proof using.
  intros. eapply local_frame_gc with (H2 := \[]); eauto; rew_heap~.
Qed.

(** Weakening on pre and post with gc-pre from [local] *)

Lemma local_weaken_gc_pre : forall B H' Q' (F:~~B) (H:hprop) Q,
  is_local F ->
  F H' Q' ->
  H ==> H' \* \GC ->
  Q' ===> Q ->
  F H Q.
Proof using.
  intros. apply* (@local_gc_pre_on (\GC) H'). hchange H2. hsimpl.
  applys* local_weaken.
Qed.

(** Weakening on pre from [local] *)

Lemma local_weaken_pre : forall H' B (F:~~B) H Q,
  is_local F ->
  F H' Q ->
  H ==> H' ->
  F H Q.
Proof using. intros. apply* local_weaken. Qed.

(** Weakening on post from [local] *)

Lemma local_weaken_post : forall B Q' (F:~~B) H Q,
  is_local F ->
  F H Q' ->
  Q' ===> Q ->
  F H Q.
Proof using. intros. apply* local_weaken. Qed.

(** Extraction of pure facts from [local] *)

Lemma local_extract_prop : forall B (F:~~B) H (P:Prop) Q,
  is_local F -> (P -> F H Q) -> F (\[P] \* H) Q.
Proof using.
  introv L M. rewrite L. intros h Hh.
  rewrite hprop_star_pure in Hh. destruct Hh as (HP&Hh).
  exists H \[] Q. splits; rew_heap~. hsimpl.
Qed.

(** Extraction of existentials from [local] *)

Lemma local_extract_exists_heap : forall B (F:~~B) A (J:A->hprop) Q,
  is_local F ->
  (forall x, F (J x) Q) ->
  F (hprop_exists J) Q.
Proof using.
  introv L M. rewrite L. introv (x&Hx).
  exists (J x) \[] Q. splits~. rew_heap~. hsimpl.
Qed.

(** Extraction of existentials below a star from [local] *)

Lemma local_extract_exists : forall B (F:~~B) H A (J:A->hprop) Q,
  is_local F ->
  (forall x, F ((J x) \* H) Q) ->
   F (hprop_exists J \* H) Q.
Proof using.
  introv L M. rewrite L. intros h Hh.
  rewrite hprop_star_exists in Hh.
  destruct Hh as (x&Hh).
  exists (J x \* H) \[] Q. splits~.
  { rew_heap~. }
  { hsimpl. }
Qed.

(** Extraction of heap representation from [local] *)

Lemma local_name_heap : forall B (F:~~B) (H:hprop) Q,
  is_local F ->
  (forall h, H h -> F (= h) Q) ->
  F H Q.
Proof using.
  introv L M. rewrite L. introv Hh. exists (= h) \[] Q. splits~.
  { rewrite hprop_star_comm. rewrite hprop_empty_def. rewrite~ hprop_star_pure. }
  { hsimpl. }
Qed.

(** Weakening under a [local] modifier *)

Lemma local_weaken_body : forall (B:Type) (F F':~~B),
  F ===> F' ->
  local F ===> local F'.
Proof using.
  introv M. intros H Q N. introv Hh. 
  destruct (N _ Hh) as (H1&H2&Q1&P1&P2&P3).
  specializes M H1 Q1. exists___*.
Qed.


(*------------------------------------------------------------------*)
(* ** Tactic [xlocal] to prove side-conditions of the form [local F]. *)

Ltac xlocal_core tt :=
  try assumption.

Tactic Notation "xlocal" :=
  xlocal_core tt.


(*------------------------------------------------------------------*)
(* ** Tactic [xpull_check] tests whether it would be useful 
      to call [xpull] to extract things from the precondition.
      Applies to a goal of the form [F H Q]. *)

Ltac xpull_check tt :=
  let H := xprecondition tt in
  hpull_check_rec H.


(*------------------------------------------------------------------*)
(* ** Tactic [xpull] to extract existentials and pure facts from
      pre-conditions. *)

(** Lemmas *)

Lemma xpull_start : forall B (F:~~B) H Q,
  is_local F -> F (\[] \* H) Q -> F H Q.
Proof using. intros. rew_heap in *. auto. Qed.

Lemma xpull_keep : forall B (F:~~B) H1 H2 H3 Q,
  F ((H2 \* H1) \* H3) Q -> F (H1 \* (H2 \* H3)) Q.
Proof using. intros. rewrite (hprop_star_comm H2) in H. rew_heap in *. auto. Qed.

Lemma xpull_assoc : forall B (F:~~B) H1 H2 H3 H4 Q,
  F (H1 \* (H2 \* (H3 \* H4))) Q -> F (H1 \* ((H2 \* H3) \* H4)) Q.
Proof using. intros. rew_heap in *. auto. Qed.

Lemma xpull_starify : forall B (F:~~B) H1 H2 Q,
  F (H1 \* (H2 \* \[])) Q -> F (H1 \* H2) Q.
Proof using. intros. rew_heap in *. auto. Qed.

Lemma xpull_empty : forall B (F:~~B) H1 H2 Q,
  is_local F -> (F (H1 \* H2) Q) -> F (H1 \* (\[] \* H2)) Q.
Proof using. intros. rew_heap. auto. Qed.

Lemma xpull_prop : forall B (F:~~B) H1 H2 (P:Prop) Q,
  is_local F -> (P -> F (H1 \* H2) Q) -> F (H1 \* (\[P] \* H2)) Q.
Proof using.
  intros. rewrite hprop_star_comm_assoc. apply~ local_extract_prop.
Qed.

Lemma xpull_id : forall A (x X : A) B (F:~~B) H1 H2 Q,
  is_local F -> (x = X -> F (H1 \* H2) Q) -> F (H1 \* (x ~> Id X \* H2)) Q.
Proof using. intros. unfold Id. apply~ xpull_prop. Qed.

Lemma xpull_exists : forall B (F:~~B) H1 H2 A (J:A->hprop) Q,
  is_local F ->
  (forall x, F (H1 \* ((J x) \* H2)) Q) ->
   F (H1 \* (hprop_exists J \* H2)) Q.
Proof using.
  intros. rewrite hprop_star_comm_assoc. apply~ local_extract_exists.
  intros. rewrite~ hprop_star_comm_assoc.
Qed.

Ltac xpull_setup tt :=
  try match goal with |- ?H ==> ?H' => 
        fail 100 "you should call hpull, not xpull" end;
  pose ltac_mark;
  apply xpull_start; [ try xlocal | ].

Ltac xpull_cleanup tt :=
  remove_empty_heaps_formula tt;
  gen_until_mark.

Ltac xpull_step tt :=
  let go HP :=
    match HP with _ \* ?HN =>
    match HN with
    | ?H \* _ =>
      match H with
      | \[] => apply xpull_empty; try xlocal
      | \[_] => apply xpull_prop; [ try xlocal | intro ]
      | hprop_exists _ => apply xpull_exists; [ try xlocal | intro ]
      | _ ~> Id _ => apply xpull_id; [ try xlocal | intro ]
      | _ \* _ => apply xpull_assoc
      | _ => apply xpull_keep
      end
    | \[] => fail 1
    | _ => apply xpull_starify
    end end in
  on_formula_pre ltac:(go).

Ltac xpull_main tt :=
  xpull_setup tt;
  (repeat (xpull_step tt));
  xpull_cleanup tt.

Tactic Notation "xpull" := xpull_main tt.
Tactic Notation "xpull" "~" := xpull; auto_tilde.
Tactic Notation "xpull" "*" := xpull; auto_star.


(*------------------------------------------------------------------*)
(* ** [xapply] and [xapplys] *)

(** [xapply E] applies a lemma [E] modulo frame/weakening.
    It applies to a goal of the form [F H Q], and replaces it
    with [F ?H' ?Q'], applies [E] to the goal, then produces
    the side condition [H ==> ?H'] and,
    - if [Q] is instantiated, then leaves [?Q' ===> Q \* \GC]
    - otherwise instantiates [Q] with [Q'].

    [xapplys E] is like [xapply E] but also attemps to simplify
    the subgoals.
*)

Ltac xapply_core H cont1 cont2 :=
  forwards_nounfold_then H ltac:(fun K =>
    match xpostcondition_is_evar tt with
    | true => eapply local_frame; [ xlocal | sapply K | cont1 tt | try apply rel_le_refl ]
    | false => eapply local_frame_gc; [ xlocal | sapply K | cont1 tt | cont2 tt ]
    end).

Ltac xapply_base H :=
  xpull_check tt;
  xapply_core H ltac:(fun tt => idtac) ltac:(fun tt => idtac).

Ltac xapplys_base H :=
  xpull_check tt;
  xapply_core H ltac:(fun tt => hcancel) ltac:(fun tt => hsimpl).

Tactic Notation "xapply" constr(H) :=
  xapply_base H.
Tactic Notation "xapply" "~" constr(H) :=
  xapply H; auto_tilde.
Tactic Notation "xapply" "*" constr(H) :=
  xapply H; auto_star.

Tactic Notation "xapplys" constr(H) :=
  xapplys_base H.
Tactic Notation "xapplys" "~" constr(H) :=
  xapplys H; auto_tilde.
Tactic Notation "xapplys" "*" constr(H) :=
  xapplys H; auto_star.


(*--------------------------------------------------------*)
(* ** [xchange] *)

(** [xchange E] applies to a goal of the form [F H Q]
    and to a lemma [E] of type [H1 ==> H2] or [H1 = H2].
    It replaces the goal with [F H' Q], where [H']
    is computed by replacing [H1] with [H2] in [H].

    The substraction is computed by solving [H ==> H1 \* ?H']
    with [hsimpl]. If you need to solve this implication by hand,
    use [xchange_no_simpl E] instead.

    [xchange <- E] is useful when [E] has type [H2 = H1]
      instead of [H1 = H2].

    [xchange_show E] is useful to visualize the instantiation
    of the lemma used to implement [xchange].
    *)

(* Lemma used by [xchange] *)

Lemma xchange_lemma : forall H1 H1' H2 B H Q (F:~~B),
  is_local F ->
  (H1 ==> H1') ->
  (H ==> H1 \* H2) ->
  F (H1' \* H2) Q ->
  F H Q.
Proof using.
  introv W1 L W2 M. applys local_frame __ \[]; eauto.
  hsimpl. hchange~ W2. hsimpl~. rew_heap~.
Qed.

Ltac xchange_apply L cont :=
   eapply xchange_lemma;
     [ try xlocal | applys L | cont tt | (*xtag_pre_post*) ].

(* Note: the term modif should be either himpl_proj1 or himpl_proj2 *)
Ltac xchange_forwards L modif cont :=
  forwards_nounfold_then L ltac:(fun K =>
  match modif with
  | __ =>
     match type of K with
     | _ = _ => xchange_apply (@himpl_proj1 _ _ _ K) cont
     | _ => xchange_apply K cont
     end
  | _ => xchange_apply (@modif _ _ _ K) cont
  end).

Ltac xchange_with_core cont H H' :=
  eapply xchange_lemma with (H1:=H) (H1':=H');
    [ try xlocal | | cont tt | (* xtag_pre_post*) ].

Ltac xchange_core cont E modif :=
  match E with
  | ?H ==> ?H' => xchange_with_core cont H H'
  | _ => xchange_forwards E modif ltac:(fun _ => instantiate; cont tt)
  end.

Ltac xchange_base cont E modif :=
  xpull_check tt;
  match goal with
  | |- _ ==> _ => hchange_base E modif
  | |- _ ===> _ => hchange_base E modif
  | _ => xchange_core cont E modif
  end.

Tactic Notation "xchange" constr(E) :=
  xchange_base ltac:(fun tt => hsimpl) E __.
Tactic Notation "xchange" "~" constr(E) :=
  xchange E; auto_tilde.
Tactic Notation "xchange" "*" constr(E) :=
  xchange E; auto_star.

Tactic Notation "xchange" "<-" constr(E) :=
  xchange_base ltac:(fun tt => hsimpl) E himpl_proj2.
Tactic Notation "xchange" "~" "<-" constr(E) :=
  xchange <- E; auto_tilde.
Tactic Notation "xchange" "*" "<-" constr(E) :=
  xchange <- E; auto_star.

Tactic Notation "xchange_no_simpl" constr(E) :=
  xchange_base ltac:(idcont) E __.
Tactic Notation "xchange_no_simpl" "<-" constr(E) :=
  xchange_base ltac:(idcont) E himpl_proj2.

Tactic Notation "xchange_show" constr(E) :=
  xchange_forwards E __ ltac:(idcont).
Tactic Notation "xchange_show" "<-" constr(E) :=
  xchange_forwards himpl_proj2 ltac:(idcont).



(*------------------------------------------------------------------*)


Global Opaque repr. 

End SepLogicSetup.




  (* TODO: fix priority of hexists vs star notation *)
